
!==============================================================================!
!==============================================================================!
!                                                                              |
!              ||===\\                                                         | 
!              ||    \\                                                        |
!              ||     ||   //==\\   ||  ||   //==||  ||/==\\                   |
!              ||     ||  ||    ||  ||  ||  ||   ||  ||    ||                  |
!              ||    //   ||    ||  ||  ||  ||   ||  ||                        |
!              ||===//     \\==//    \\==\\  \\==\\  ||                        |
!                                                                              |
!==============================================================================!
!==============================================================================!
!                                                                              |
!              TRACKING CLOUD POST PROCESSING   Jan. 2015                      |
!              written by Matthias Schmiddunser                                |
!              using age calculation form Pecube and Terra                     |
!                                                                              |
!==============================================================================!
!==============================================================================!
!                                                                              |
! This routine reads in the tracking output files, extracts the surfaced       |
! tracks, does variable calculation and writes ASCII vtk files.                |
!                                                                              |
! CAUTION: Particle ID (Douar) and Point Number (VTK) differ: DOUAR numbering  |
!     starts at 1, VTK numbering starts at 0.                                  |

program post_tracking

use trdefinitions
use m_ageCalculation

implicit none

include 'mpif.h'

type (tracking_cloud) :: trcl
type (input_parameters) :: inparam
type (output_parameters) :: outparam
 
!buffer variables for reading
real :: read_time,lref,veloref,rhoref,tempscale
integer :: read_step,read_ntracks
integer,dimension(:,:),allocatable :: read_tracks
integer,dimension(:),allocatable :: keep_tracks

!mpi and data transfer variables
logical :: flag
integer :: provided,iproc,nproc,ierr
integer (kind=MPI_ADDRESS_KIND) :: maxtag
integer,dimension(MPI_STATUS_SIZE) :: mpistatus
integer :: sender,tracksent,send_nstep,endstep
real,dimension(:),allocatable :: send_time,send_temp
real,dimension(:),allocatable :: allages

!age calculation
real(8) :: apatiteHe, zirconHe
real(8) :: zirconFT, apatiteFT
real(8) :: muscovite_age


!other variables
integer :: i,k,l,err,dummyint,itrack
real :: dummyreal
logical :: keepme

character(4) :: cistep

!------------------------------------------------------------------------------|
!------------------------------------------------------------------------------|

provided=0
call mpi_init_thread(MPI_THREAD_MULTIPLE,provided,ierr)
call mpi_comm_size (mpi_comm_world,nproc,ierr)
call mpi_comm_rank (mpi_comm_world,iproc,ierr)


if (iproc==0) then
  write (*,'(a)') '!------------------------------------------------------------------------------|'
  write (*,'(a)') '!------------------------------------------------------------------------------|'
  write (*,'(a)') '!                                                                              |'
  write (*,'(a)') '!          8888888b.   .d88888b.  888     888       d8888 8888888b.            |'
  write (*,'(a)') '!          888  "Y88b d88P" "Y88b 888     888      d88888 888   Y88b           |'
  write (*,'(a)') '!          888    888 888     888 888     888     d88P888 888    888           |'
  write (*,'(a)') '!          888    888 888     888 888     888    d88P 888 888   d88P           |'
  write (*,'(a)') '!          888    888 888     888 888     888   d88P  888 8888888P"            |'
  write (*,'(a)') '!          888    888 888     888 888     888  d88P   888 888 T88b             |'
  write (*,'(a)') '!          888  .d88P Y88b. .d88P Y88b. .d88P d8888888888 888  T88b            |'
  write (*,'(a)') '!          8888888P"   "Y88888P"   "Y88888P" d88P     888 888   T88b           |'
  write (*,'(a)') '!                                                                              |'
  write (*,'(a)') '!             DOUAR-WSMP post-processor for the tracking cloud                 |'
  write (*,'(a)') '!------------------------------------------------------------------------------|'
  write (*,'(a)') '!------------------------------------------------------------------------------|'
  write (*,*) ''
end if

!------------------------------------------------------------------------------|
!    Input parameters
!------------------------------------------------------------------------------|

!Reading input parameters from post-input-file:

if (iproc==0) then
    open(unit=10, file='input_tracking_outputs.txt',status='old')

    read (10,*) inparam%dir
    read (10,*) outparam%dir
    read (10,*) outparam%mode
    read (10,*) outparam%scaled
    read (10,*) outparam%first
    read (10,*) outparam%last
    read (10,*) outparam%step
    read (10,*) outparam%keep_nstep
    read (10,*) outparam%min_track
    !outparam%keep_nstep has to be less than outparam%step
    outparam%keep_nstep = min(outparam%keep_nstep,outparam%step-1)
    read (10,*) inparam%first
    read (10,*) inparam%step
    read (10,*)  !here come the age parameters
    read (10,*) outparam%pad_age
    read (10,*) outparam%AHe
    read (10,*) outparam%ZHe
    read (10,*) outparam%AFT
    read (10,*) outparam%ZFT
    read (10,*) outparam%Muscovite
    read (10,*) outparam%ageID
    read (10,*)  !here come the track parameters
    read (10,*) outparam%velo
    read (10,*) outparam%veloz
    read (10,*) outparam%temp
    read (10,*) outparam%press
    read (10,*) outparam%time
    read (10,*) outparam%istep
    read (10,*) outparam%trackID

    close (10)

    !calculating and andjusting input and output timesteps
    outparam%nstep = (outparam%last-outparam%first)/outparam%step+1
    outparam%last = outparam%first+(outparam%nstep-1)*outparam%step

    inparam%last = outparam%last
    inparam%nstep = (inparam%last-inparam%first)/inparam%step+1
    inparam%last = inparam%first+(inparam%nstep-1)*inparam%step

    !Checking for slash at end of input and output dir
    l = len(trim(inparam%dir))
    if (inparam%dir(l:l) .ne. '/') then
        inparam%dir = inparam%dir(1:l)//'/'
    end if
    l = len(trim(outparam%dir))
    if (outparam%dir(l:l) .ne. '/') then
        outparam%dir = outparam%dir(1:l)//'/'
    end if
end if

!Distribute output parameters to other processes
call mpi_bcast(outparam%mode,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast(outparam%AHe,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast(outparam%ZHe,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast(outparam%AFT,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast(outparam%ZFT,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast(outparam%Muscovite,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast(outparam%ageID,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast(outparam%velo,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast(outparam%veloz,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast(outparam%temp,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast(outparam%press,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast(outparam%time,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast(outparam%istep,1,mpi_integer,0,mpi_comm_world,ierr)
call mpi_bcast(outparam%trackID,1,mpi_integer,0,mpi_comm_world,ierr)

!determine number of ages that will be calculated
outparam%nages = outparam%AHe+outparam%ZHe+outparam%AFT+outparam%ZFT+outparam%Muscovite
allocate (allages(outparam%nages),stat=err); if (err.ne.0) call stop_run ('Error alloc allages')

!writing output parameters to screen
if (iproc==0) then
    write (*,'(a,a)') 'Simulation Directory: ',inparam%dir
    write (*,'(a,i2,a,i3,a,i3,a,i3)') 'Input Timesteps: Reading ',inparam%nstep,' input files '// & 
    'from step ',inparam%first,' to ',inparam%last,' in steps of ',inparam%step
    write (*,'(a,a)') 'Output Directory: ',outparam%dir
    write (*,'(a,i2,a,i3,a,i3,a,i3)') 'Output Timesteps: Generating ',outparam%nstep,' output files '// &
    'from step ',outparam%first,' to ',outparam%last,' in steps of ',outparam%step
    write (*,'(a,i2,a)') 'accumulating ',outparam%keep_nstep,' steps for surfaced tracks'
    write (*,'(a,i1)') 'Output mode: ',outparam%mode
    write (*,'(a,l)') 'Write natural units:     ',(outparam%scaled==1)
    write (*,'(a)') ''
    write (*,'(a)') ' Age variables: '
    write (*,'(a,l)') 'Write AHe Ages:          ',(outparam%AHe==1)
    write (*,'(a,l)') 'Write ZHe Ages:          ',(outparam%ZHe==1)
    write (*,'(a,l)') 'Write AFT Ages:          ',(outparam%AFT==1)
    write (*,'(a,l)') 'Write ZFT Ages:          ',(outparam%ZFT==1)
    write (*,'(a,l)') 'Write Muscovite Ages:    ',(outparam%Muscovite==1)
    write (*,'(a,l)') 'Write track ID           ',(outparam%ageID==1)
    write (*,'(a)') ''
    write (*,'(a)') 'Track variables: '
    write (*,'(a,l)') 'Write track velocity:    ',(outparam%velo==1)
    write (*,'(a,l)') 'Write track z velocity:  ',(outparam%veloz==1)
    write (*,'(a,l)') 'Write track temperature: ',(outparam%temp==1)
    write (*,'(a,l)') 'Write track pressure:    ',(outparam%press==1)
    write (*,'(a,l)') 'Write track time:        ',(outparam%time==1)
    write (*,'(a,l)') 'Write track step number: ',(outparam%istep==1)
    write (*,'(a,l)') 'Write particle/track ID: ',(outparam%trackID==1)
    write (*,'(a)') ''
    write (*,'(a)') '!------------------------------------------------------------------------------|'
    write (*,'(a)') ''
end if

!------------------------------------------------------------------------------|
!    Reading Tracking Cloud Files
!------------------------------------------------------------------------------|

if (iproc==0) then
    !reading input file header for array sizes
    call int_to_char (cistep,4,inparam%last)

    open (unit=20,file=trim(inparam%dir)//'TROUT/tracking_t'//cistep//'.bin', &
           status='old',form='unformatted')
    read (20) trcl%np,read_ntracks,read_step,read_time
    read (20) lref,veloref,rhoref,tempscale
    close (20)

    !allocating master arrays for input
    allocate (trcl%x(trcl%np,inparam%nstep),stat=err)
    if (err.ne.0) call stop_run ('Error alloc trcl%x')
    allocate (trcl%y(trcl%np,inparam%nstep),stat=err)
    if (err.ne.0) call stop_run ('Error alloc trcl%y')
    allocate (trcl%z(trcl%np,inparam%nstep),stat=err)
    if (err.ne.0) call stop_run ('Error alloc trcl%z')
    allocate (trcl%temp(trcl%np,inparam%nstep),stat=err)
    if (err.ne.0) call stop_run ('Error alloc trcl%temp')
    allocate (trcl%press(trcl%np,inparam%nstep),stat=err)
    if (err.ne.0) call stop_run ('Error alloc trcl%press')
    allocate (trcl%step(inparam%nstep),stat=err)
    if (err.ne.0) call stop_run ('Error alloc trcl%step')
    allocate (trcl%time(inparam%nstep),stat=err)
    if (err.ne.0) call stop_run ('Error alloc trcl%time')
    allocate (read_tracks(read_ntracks,3),stat=err)
    if (err.ne.0) call stop_run ('Error alloc read_tracks')

    !read all input files and save in collective arrays
    k = 1
    do i = inparam%first,inparam%last,inparam%step
        !read all information for this time step
        call int_to_char (cistep,4,i)
        open (unit=20,file=trim(inparam%dir)//'TROUT/tracking_t'//cistep//'.bin', & 
               status='old',form='unformatted')
        
        read (20) dummyint,dummyint,trcl%step(k),trcl%time(k)
        read (20) 
        read (20) trcl%x(:,k)
        read (20) trcl%y(:,k)
        read (20) trcl%z(:,k)
        read (20) trcl%temp(:,k)
        read (20) trcl%press(:,k)
        
        !read track information only from last timestep
        if (i==inparam%last) then
            write (*,'(a)') 'All surfaced tracks:'
            write (*,'(a)') 'track ID, Particle ID, start, end'
            do l=1,read_ntracks 
                read (20) read_tracks(l,:)
                write (*,'(4i6)') l, read_tracks(l,:)
            end do
            write (*,'(a)') ''
        end if

        !advance column in master array
        k = k+1

        close (20)
    end do

    write (*,'(a)') 'Finished reading input files'
    write (*,'(a)') ''

    !select only tracks that surface at an output timestep or up to outparam%keep_nstep
    !timesteps before each output timestep and are at least min_track_lenght 
    !timesteps long.
    !Then adjust track timesteps to actual position in master array.
    !The offset k has to be recorded in order to associate the track with the
    !correct timestep.

    allocate (keep_tracks(read_ntracks),stat=err)
    if (err.ne.0) call stop_run ('Error alloc keep_tracks')
    keep_tracks = 0
    trcl%ntracks = 0
    do k=0,outparam%keep_nstep
        do i=1,read_ntracks
            keepme = read_tracks(i,3) >= outparam%first-k
            keepme = keepme .AND. (mod((read_tracks(i,3)+k-outparam%first),outparam%step)==0)
            keepme = keepme .AND. read_tracks(i,3)-read_tracks(i,2)+1 >= outparam%min_track
            if (keepme) then
                keep_tracks(i) = 1000+k
                trcl%ntracks = trcl%ntracks+1
            end if
        end do
    end do

    !allocate tracks array of needed length

    if (trcl%ntracks > 0) then
        allocate (trcl%tracks(trcl%ntracks,7),stat=err)
        if (err.ne.0) call stop_run ('Error alloc trcl%tracks')
        trcl%tracks = 0

        !copy relevant tracks into new track array
        !adjust track timsteps to position in new track array
        !always round up for start and round down for end in order not to 
        !include parts of other tracks
        itrack=1
        write (*,'(a)') 'Selected surfaced tracks:'
        write (*,'(a)') '*: in master array'
        write (*,'(a)') 'Particle ID, start*, end*, length*, output timestep, output step*, track ID'
        do i=1,read_ntracks
            if (keep_tracks(i)>0) then
                !1st column is particle ID
                trcl%tracks(itrack,1) = read_tracks(i,1)
                !2nd column is start in master array, round up
                trcl%tracks(itrack,2)= max(1,ceiling(1.0+(read_tracks(i,2)-inparam%first)/(1.0*inparam%step)))
                !3rd columd is end in master array, round down
                trcl%tracks(itrack,3)= max(1,1+(read_tracks(i,3)-inparam%first)/inparam%step) !integer division
                !4th column is track length (number of points in master array)
                trcl%tracks(itrack,4) = trcl%tracks(itrack,3)-trcl%tracks(itrack,2)+1
                !5th column is Douar timestep of surfacing, upshifted for those tracks 
                !within outparam%keep_nstep
                trcl%tracks(itrack,5) = read_tracks(i,3)+(keep_tracks(i)-1000)
                !6th column is output timestep in master array
                trcl%tracks(itrack,6)= max(1,1+(trcl%tracks(itrack,5)-inparam%first)/inparam%step) !integer division
                !7th column is track ID, i.e. the row of the original track array from input file
                trcl%tracks(itrack,7) = i

                write (*,'(7i6)') trcl%tracks(itrack,:)
                itrack = itrack+1
            end if
        end do
    else
        write (*,'(a)') 'No surfaced tracks'
    end if

    deallocate (read_tracks)
    deallocate (keep_tracks)
end if

call mpi_bcast(trcl%ntracks,1,mpi_integer,0,mpi_comm_world,ierr)

!------------------------------------------------------------------------------|
!    Calculate Thermochron Ages
!------------------------------------------------------------------------------|
ages: if ((outparam%mode==1 .or. outparam%mode==2) .AND. trcl%ntracks>0) then
    multiproc: if (nproc > 1) then
        !Check for enough tags for self scheduling
        call MPI_COMM_GET_ATTR (MPI_COMM_WORLD, MPI_TAG_UB, maxtag, flag, ierr)
        if (int(trcl%ntracks,MPI_ADDRESS_KIND) > maxtag) then
            call stop_run ('Too many tracks for safe communications')
        end if

        workshare: if (iproc==0) then
        !master distributing work
            if (outparam%AHe==1) allocate (trcl%AHe(trcl%ntracks))
            if (outparam%ZHe==1) allocate (trcl%ZHe(trcl%ntracks))
            if (outparam%AFT==1) allocate (trcl%AFT(trcl%ntracks))
            if (outparam%ZFT==1) allocate (trcl%ZFT(trcl%ntracks))
            if (outparam%Muscovite==1) allocate (trcl%Muscovite(trcl%ntracks))

            tracksent = 0
            !write (* ,*) 'iproc',iproc , 'ntracks:', trcl%ntracks, 'nproc: ',nproc, 'tracksent', tracksent
            !wait for answer, distribute new track
            workdistr: do i=2-nproc,trcl%ntracks
                !for f irst nproc-1 times, send to processes without receive
                !distribute one track to each processor

                if (i>0) then
                    !get resul t from other process
                    
                    call MPI_RECV (allages,outparam%nages,MPI_REAL,MPI_ANY_SOURCE, & 
                                    MPI_ANY_TAG,MPI_COMM_WORLD,mpistatus,ierr)
                    !Identify process and track ID
                    sender = mpistatus(MPI_SOURCE)
                    itrack = mpistatus(MPI_TAG)

                    write (*,'(i3,a,i5,a,i3,i2)') iproc, ': received track',itrack,' from proc ',sender, 'length = ',outparam%nages

                    !Write into result arrays
                    k = 1
                    if (outparam%AHe==1) then
                        trcl%AHe(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%ZHe==1) then
                        trcl%ZHe(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%AFT==1) then
                        trcl%AFT(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%ZFT==1) then
                        trcl%ZFT(itrack) = allages(k)
                        k = k+1
                    end if
                    if (outparam%Muscovite==1) then
                        trcl%Muscovite(itrack) = allages(k)
                    end if
                else
                    !effecively loops from 1 to nproc-1 for sender
                    sender = i+nproc-1
                end if

                !Send new information
                if (tracksent < trcl%ntracks) then
                    !Length of time and temperature arrays 
                    !Add two entrys at front for initalisation
                    !If surfaced earlier than output timestep, add another entry at end
                    endstep = min(1,trcl%tracks(tracksent+1,6)-trcl%tracks(tracksent+1,3)) 
                    send_nstep = 2 + trcl%tracks(tracksent+1,4)+endstep

                    allocate (send_time(send_nstep))
                    allocate (send_temp(send_nstep))

                    !pad send-arrays at beginning, part 1
                    send_time(3:send_nstep-endstep) = &
                        trcl%time(trcl%tracks(tracksent+1,2):trcl%tracks(tracksent+1,3))
                    send_time(1:2) = (/0.0,0.0/)

                    send_temp(3:send_nstep-endstep) = trcl%temp(trcl%tracks(tracksent+1,1), &
                        trcl%tracks(tracksent+1,2):trcl%tracks(tracksent+1,3))
                    send_temp(1) = send_temp(3)
                    send_temp(2) = send_temp(3)
                    
                    !pad send arrays at end if necessary
                    if (endstep>0) then
                        send_time(send_nstep) = trcl%time(trcl%tracks(tracksent+1,6))
                        send_temp(send_nstep) = send_temp(send_nstep-1)
                    end if

                    do l=1,send_nstep
                        !Calculate relative Age in Ma
                        send_time(l) = send_time(send_nstep) - send_time(l)
                        send_time(l) = send_time(l)*lref/veloref 
                        !scale temperature
                        send_temp(l) = send_temp(l)*tempscale
                    end do

                    !pad age array at the beginning, part 2
                    send_time(1) = outparam%pad_age
                    send_time(2) = 0.5*(send_time(1)+send_time(3))

                    !write (*,*) iproc, ': send track',tracksent+1,' of length',send_nstep,' to proc ',sender
                    call MPI_SEND (send_time,send_nstep,MPI_REAL,sender,tracksent+1, & 
                                    MPI_COMM_WORLD,ierr)
                    call MPI_SEND (send_temp,send_nstep,MPI_REAL,sender,tracksent+1, & 
                                    MPI_COMM_WORLD,ierr)
                    tracksent = tracksent+1

                    deallocate (send_time)
                    deallocate (send_temp)

                else
                    !send end signal
                    call MPI_SEND (0.0,1,MPI_REAL,sender,0,MPI_COMM_WORLD,ierr)
                end if
            end do workdistr


        else workshare
        !slaves doing actual calculation
            !write (* ,*) 'iproc',iproc , 'ntracks:', trcl%ntracks, 'nproc: ',nproc
            agecalc: do
                call MPI_probe  (0, MPI_ANY_TAG, MPI_COMM_WORLD, mpistatus, ierr)
                itrack = mpistatus(MPI_TAG)
                !write (*,*) iproc, ': probed track', itrack

                !tag 0 is end signal
                if (itrack==0) then
                    !write (*,*) iproc, ': receive end signal'
                    !empty send buffer - necessary?
                    call MPI_RECV (dummyreal,1,MPI_REAL,0,MPI_ANY_TAG,MPI_COMM_WORLD,mpistatus,ierr)
                    exit
                end if

                call MPI_get_count (mpistatus,MPI_REAL,send_nstep,ierr)

                allocate (send_time(send_nstep))
                allocate (send_temp(send_nstep))

                !write (*,*) iproc, ': receiving track', itrack, ' of length', send_nstep
                call MPI_RECV (send_time,send_nstep,MPI_REAL,0, & 
                                MPI_ANY_TAG,MPI_COMM_WORLD,mpistatus,ierr)
                call MPI_RECV (send_temp,send_nstep,MPI_REAL,0, & 
                                MPI_ANY_TAG,MPI_COMM_WORLD,mpistatus,ierr)

                !do age calculation
                k=1
                if (outparam%AHe==1 .or. outparam%ZHe==1) then
                    call Mad_He(dble(send_time), dble(send_temp), send_nstep, apatiteHe, zirconHe)
                    !apatiteHe = 1.d0
                    !zirconHe = 2.d0
                    if (outparam%AHe==1) then
                        call correct_number(apatiteHe)
                        allages(k) = real(apatiteHe)
                        k = k+1
                    end if
                    if (outparam%ZHe==1)  then
                        call correct_number(zirconHe)
                        allages(k) = real(zirconHe)
                        k = k+1
                    end if
                end if

                if (outparam%AFT==1) then
                    call Mad_Trax_AFT (dble(send_time), dble(send_temp), send_nstep, apatiteFT)
                    !call get_aft_age(dble(send_time), dble(send_temp), send_nstep, apatiteFT)
                    !apatiteFT = 3.d0
                    call correct_number(apatiteFT)
                    allages(k) = real(apatiteFT)
                    k = k+1
                end if

                if (outparam%ZFT==1) then
                    call ZFT(dble(send_time), dble(send_temp), send_nstep, zirconFT)
                    !zirconFT = 4.d0
                    call correct_number(zirconFT)
                    allages(k) = real(zirconFT)
                    k = k+1
                end if

                if (outparam%Muscovite==1) then
                    call Muscovite(dble(send_time), dble(send_temp), send_nstep, muscovite_age)
                    !muscovite_age = 5.d0
                    call correct_number(muscovite_age)
                    allages(k) = real(muscovite_age)
                    k = k+1
                end if

                !send results
                !write (*,*) iproc, ': send results of track', itrack, 'length = ',outparam%nages
                call MPI_SEND (allages,outparam%nages,MPI_REAL,0,itrack, MPI_COMM_WORLD,ierr)
                 
                deallocate (send_time)
                deallocate (send_temp)

            end do agecalc
        end if workshare
    else multiproc
        !Do calculation only on one processor
        !allocate age arrays
        if (outparam%AHe==1) allocate (trcl%AHe(trcl%ntracks))
        if (outparam%ZHe==1) allocate (trcl%ZHe(trcl%ntracks))
        if (outparam%AFT==1) allocate (trcl%AFT(trcl%ntracks))
        if (outparam%ZFT==1) allocate (trcl%ZFT(trcl%ntracks))
        if (outparam%Muscovite==1) allocate (trcl%Muscovite(trcl%ntracks))

        do i = 1,trcl%ntracks
            !calculate temp and time array lenghts for given track
            endstep = min(1,trcl%tracks(i,6)-trcl%tracks(i,3)) 
            send_nstep = 2 + trcl%tracks(i,4)+endstep

            allocate (send_time(send_nstep))
            allocate (send_temp(send_nstep))

            !pad send-arrays at beginning, part 1
            send_time(3:send_nstep-endstep) = &
                trcl%time(trcl%tracks(i,2):trcl%tracks(i,3))
            send_time(1:2) = (/0.0,0.0/)

            send_temp(3:send_nstep-endstep) = trcl%temp(trcl%tracks(i,1), &
                trcl%tracks(i,2):trcl%tracks(i,3))
            send_temp(1) = send_temp(3)
            send_temp(2) = send_temp(3)
            
            !pad send arrays at end if necessary
            if (endstep>0) then
                send_time(send_nstep) = trcl%time(trcl%tracks(i,6))
                send_temp(send_nstep) = send_temp(send_nstep-1)
            end if

            do l=1,send_nstep
                !Calculate relative Age in Ma
                send_time(l) = send_time(send_nstep) - send_time(l)
                send_time(l) = send_time(l)*lref/veloref 
                !scale temperature
                send_temp(l) = send_temp(l)*tempscale
            end do

            !pad age array at the beginning, part 2
            send_time(1) = outparam%pad_age
            send_time(2) = 0.5*(send_time(1)+send_time(3))

            if (outparam%AHe==1 .OR. outparam%ZHe==1) then
                call Mad_He(dble(send_time), dble(send_temp), send_nstep, apatiteHe, zirconHe)
                !apatiteHe = 1.d0
                !zirconHe = 2.d0
                if (outparam%AHe==1) then
                    call correct_number(apatiteHe)
                    trcl%AHe(i) = real(apatiteHe)
                end if
                if (outparam%ZHe==1) then
                    call correct_number(zirconHe)
                    trcl%ZHe(i) = real(zirconHe)
                end if
            end if

            if (outparam%AFT==1) then
                call Mad_Trax_AFT (dble(send_time), dble(send_temp), send_nstep, apatiteFT)
                !call get_aft_age(dble(send_time), dble(send_temp), send_nstep, apatiteFT)
                !apatiteFT = 3.d0
                call correct_number(apatiteFT)
                trcl%AFT(i) = real(apatiteFT)
            end if

            if (outparam%ZFT==1) then
                call ZFT(dble(send_time), dble(send_temp), send_nstep, zirconFT)
                !zirconFT = 4.d0
                call correct_number(zirconFT)
                trcl%ZFT(i) = real(zirconFT)
            end if


            if (outparam%Muscovite==1) then
                call Muscovite(dble(send_time), dble(send_temp), send_nstep, muscovite_age)
                !muscovite_age = 5.d0
                call correct_number(muscovite_age)
                trcl%Muscovite(i) = real(muscovite_age)
            end if
             
            deallocate (send_time)
            deallocate (send_temp)
        end do
    end if multiproc
end if ages

!------------------------------------------------------------------------------|
!    Scaling Values to natural units
!------------------------------------------------------------------------------|
!check what age prediction needs...
!units of scaling parameters:
!lref: km
!veloref: mm/yr = km/Myr
!rhoref: kg/m^3
!tempscale: °C

if (iproc==0) then
    if (outparam%scaled==1) then
        do k=1,inparam%nstep 
            do i=1,trcl%np
                trcl%x(i,k) = trcl%x(i,k)*lref !pos in km
                trcl%y(i,k) = trcl%y(i,k)*lref
                trcl%z(i,k) = trcl%z(i,k)*lref
                trcl%temp(i,k) = trcl%temp(i,k)*tempscale !temp in C
                trcl%press(i,k) = trcl%press(i,k)*1.0e-3*9.81*rhoref*lref !pressure in MPa
            end do
            trcl%time(k) = trcl%time(k)*lref/veloref !time in Myr
        end do
    end if
end if


!------------------------------------------------------------------------------|
!    Calculating Track Output values
!------------------------------------------------------------------------------|

if (iproc==0) then
    !track values
    velo: if (outparam%velo==1) then
        allocate (trcl%velo(trcl%np,inparam%nstep),stat=err); if (err.ne.0) call stop_run ('Error alloc trcl%velo')
        do i=1,trcl%np
                trcl%velo(i,1) = sqrt( & 
                    (trcl%x(i,2)-trcl%x(i,1))**2+(trcl%y(i,2)-trcl%y(i,1))**2+(trcl%z(i,2)-trcl%z(i,1))**2) &
                    /(trcl%time(2)-trcl%time(1))
        end do
        do i=1,trcl%np
            do k=2,inparam%nstep
                trcl%velo(i,k) = sqrt( & 
                    (trcl%x(i,k)-trcl%x(i,k-1))**2+(trcl%y(i,k)-trcl%y(i,k-1))**2+(trcl%z(i,k)-trcl%z(i,k-1))**2) &
                    /(trcl%time(k)-trcl%time(k-1))
            end do
        end do
    end if velo

    veloz: if (outparam%veloz==1) then
        allocate (trcl%veloz(trcl%np,inparam%nstep),stat=err); if (err.ne.0) call stop_run ('Error alloc trcl%veloz')
        do i=1,trcl%np
            trcl%veloz(i,1) = (trcl%z(i,2)-trcl%z(i,1))/(trcl%time(2)-trcl%time(1))
        end do
        do i=1,trcl%np
            do k=2,inparam%nstep
                trcl%veloz(i,k) = (trcl%z(i,k)-trcl%z(i,k-1))/(trcl%time(k)-trcl%time(k-1))
            end do
        end do
    end if veloz
end if


!------------------------------------------------------------------------------|
!    Checking for NaNs and variable ranges
!------------------------------------------------------------------------------|


if (iproc==0) then
    do k=1,inparam%nstep
        call correct_number(trcl%time(k))
        do i=1,trcl%np
            call correct_number(trcl%temp(i,k))
            call correct_number(trcl%press(i,k))
            if (outparam%velo==1) call correct_number(trcl%velo(i,k))
            if (outparam%veloz==1) call correct_number(trcl%veloz(i,k))
        end do
    end do
end if

!------------------------------------------------------------------------------|
!    Writing Output Files
!------------------------------------------------------------------------------|

if (iproc==0)  call write_post_track (trcl, inparam, outparam)

!------------------------------------------------------------------------------|
!    Deallocate arrays
!------------------------------------------------------------------------------|

if (iproc==0) then
    deallocate (trcl%x)   
    deallocate (trcl%y)
    deallocate (trcl%z)
    deallocate (trcl%temp)
    deallocate (trcl%press)
    deallocate (trcl%step)
    deallocate (trcl%time)
    if (trcl%ntracks>0) then
        deallocate (trcl%tracks)
        if (outparam%mode==1 .or. outparam%mode==2) then
            if (outparam%AHe==1) deallocate (trcl%AHe)
            if (outparam%ZHe==1) deallocate (trcl%ZHe)
            if (outparam%AFT==1) deallocate (trcl%AFT)
            if (outparam%ZFT==1) deallocate (trcl%ZFT)
            if (outparam%Muscovite==1) deallocate (trcl%Muscovite)
        end if
    end if
    if (outparam%velo==1) deallocate (trcl%velo)
    if (outparam%veloz==1) deallocate (trcl%veloz)
end if

deallocate (allages)

call mpi_finalize (ierr)

end program post_tracking

